# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.



# Replacing tmart_class

# TMart: Topography-adjusted Monte-Carlo Adjacency-effect Radiative Transfer code

from multiprocessing import cpu_count
from pathos.multiprocessing import ProcessingPool
import numpy as np
import time
import sys

# tmart dependencies 
from .tm_geometry import dirP_to_coord 
from .tm_intersect import intersect_line_DEMtri2
from .tm_water import find_R_wc, RefraIdx
from .Tmart2 import Tmart2


# Track progress in multiprocessing
def _track_job(job, update_interval=2):
    while job._number_left > 0:
        print("Jobs remaining = {0}".format(job._number_left * job._chunksize))
        time.sleep(update_interval)


# The main object in TMart
class Tmart(Tmart2):
    '''Create a Tmart object that does radiative transfer modelling. 
    
    Arguments:

    * ``Surface`` -- Surface object from the Surface module.
    * ``Atmosphere`` -- Atmosphere object from the Atmosphere module.
    * ``shadow`` -- True or False, whether to test if light is blocked by another surface at a collision. 
    * ``VROOM`` -- VROOM acceleration. 0: no acceleration. 1: all collisions are directed towards the sun. 

    Example usage::

      my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm)

    '''
    
    def __init__(self, Surface, Atmosphere, shadow = False, VROOM = 0):

        self.Surface = Surface
        self.Atmosphere = Atmosphere
        self.shadow = shadow
        self.VROOM = VROOM 
        
        self.sensor_coords = None
        self.sun_dir = None
        self.print_on = False # print switch 
        self.plot_on = False  # don't turn it on for multiprocessing 
        
        self.target_pt_direction = [180,0] 
        self.pixel = None
        self.pixel_elevation = None
        
        # Atmosphere
        self.wl = None
        self.atm_profile_wl = None # single wavelength 
        self.aerosol_SPF_wl = None 
        
        # Wind
        self.wind_speed = 3 # default 3 m/s
        self.wind_azi_avg = True # azimuthally averaged cox munk
        self.wind_dir = 0 # default azimuthal, 0 means UPWIND along x-axis
            # AKA the direction where wind comes from 
           
        # Water
        self.F_wc_wl = None # fraction of sea surface covered by whitecaps 
        self.R_wc_wl = None # whitecap reflectance at this wavelength
        
        self.water_salinity = 0
        self.water_temperature = 25      
        self.water_refraIdx_wl = None # refractive index of water at this wavelength 
        
        # In development 
        self.output_flux = False # output irradiance reflectance, direct irradiance and diffuse irradiance on the ground, under development 
        
        
             
        
    def set_geometry(self, sun_dir=[0,0], target_pt_direction=None, sensor_coords=None, pixel=None, target_coords=None):
        '''Set the sun-target-sensor geometry. Only one of ``sensor_coords``, ``pixel`` and ``target_coords`` is needed. 
        
        Arguments:
            
        * ``sun_dir`` -- Solar angle, in [Zenith, Azimuth], relative to the target.
        * ``target_pt_direction`` -- Photon's initial moving direction', AKA viewing angle, in [Zenith, Azimuth], relative to the sensor. 
            * Alternative input: 'lambertian_up' or 'lambertian_down'. This can only be used with ``sensor_coords``. The initial direction will be a random direction up following a Lambertian distribution for each of the photons. This is used in calculating irradiance.  
        * ``sensor_coords`` -- Where the sensor is, in [X, Y, Z], unit in meters.
        * ``pixel`` -- The target pixel to shoot photons. Parallel light rays will hit random points within the square pixel. 
        * ``target_coords`` -- Where the photon will land on the surface, only [X, Y] needed, [Z] is automatically adjusted. 


        Example usage::

          my_tmart.set_geometry(sensor_coords=[51,50,130_000], 
                                target_pt_direction=[180,0],
                                sun_dir=[0,0])
          
          # sensor_coords can be replaced by pixel or target_coords
          
          my_tmart.set_geometry(target_pt_direction=[170,0],
                                target_coords=[15,10], 
                                sun_dir=[0,0])        

          my_tmart.set_geometry(target_pt_direction=[170,0],
                                pixel=[1,1], 
                                sun_dir=[0,0])     


        '''
        
        
        n_not_none = (sum(x is not None for x in [sensor_coords,pixel,target_coords]))
        
        if n_not_none != 1: sys.exit('only one of sensor_coords,pixel,target_coords should be provided')
        
        
        
        ### Sensor coordinates, take one of the three inputs 
        
        # direct input 
        if sensor_coords is not None:
            # make sure not to hit the boundary between triangles 
            if sensor_coords[0]==sensor_coords[1]: sensor_coords[1]=sensor_coords[1]+0.0001
            self.sensor_coords = np.array(sensor_coords)          
            
        # pixel based, assume height 120km 
        elif pixel is not None:          
            self.pixel_elevation = self.Surface.DEM[pixel[0],pixel[1]]
            
            # distance from target to sensor 
            # it's negative because target_pt_direction is larger than 90
            dist_120000 = (120_000 - self.pixel_elevation) / np.cos(target_pt_direction[0]/180*np.pi) 
            self.sensor_coords = dirP_to_coord(dist_120000, target_pt_direction)
            self.pixel = pixel
            
            
        # target_coords, assume height 120km
        elif target_coords is not None:      
            if target_coords[0]==target_coords[1]: target_coords[1]=target_coords[1]+0.0001
            
            # target_coords is 2d, target_coords3d includes elevation
            target_coords3d = intersect_line_DEMtri2(np.array(target_coords + [120_000]), 
                                                      np.array(target_coords + [0]), 
                                                      self.Surface.DEM_triangulated)
            
            # If there is triangle intersection 
            if target_coords3d.shape[0] > 0:
                
                # closest intersection # There seems to be only 1?
                target_coords3d_chosen = target_coords3d.iloc[target_coords3d.linear_distance.idxmin()] 
                
                # print(target_coords3d)
                q_collision = target_coords3d_chosen.tolist()[0:3]  
                q_elevation = q_collision[2]
   
            # Background collision
            else:
                q_collision = np.array(target_coords + [self.Surface.bg_elevation])  
                q_elevation = q_collision[2]
            
            dist_120000 = (120_000 - q_elevation) / np.cos(target_pt_direction[0]/180*np.pi) 
            self.sensor_coords = dirP_to_coord(dist_120000, target_pt_direction) + q_collision

        else: 
            sys.exit('only one of sensor_coords,pixel,target_coords should be provided')
            
            
        # Lock photon initial direction 
        self.target_pt_direction = target_pt_direction   
        
        # Sun direction 
        self.sun_dir = sun_dir # [zenith, azimuthal]        
        
        
    def set_wind(self,wind_speed=3,wind_azi_avg=True,wind_dir=0): 
        '''Set wind speed and direction. 
        
        Arguments:

        * ``wind_speed`` -- wind speed in meters, default 10.
        * ``wind_azi_avg`` -- cox-munk slopes azimuthally averaged, default True.
        * ``wind_dir`` -- upwind direction clockwise from east, default 0. Meaningless when wind_azi_avg is True.

        Example usage::
          
          my_tmart.set_wind(wind_speed=5)
          my_tmart.set_wind(wind_speed=5, wind_azi_avg = False, wind_dir=0)
          my_tmart.set_wind(wind_speed=5, wind_azi_avg = True)
          
        '''
        
        self.wind_speed = wind_speed
        self.wind_dir = wind_dir
        self.wind_azi_avg = wind_azi_avg
        
        
    def set_water(self,water_salinity=0,water_temperature=25): # default 0/1000 and 25C
        '''Set water salinity and temperature. 
        
        Arguments:

        * ``water_salinity`` -- water salinity in parts per thousand, default 0.
        * ``water_temperature`` -- water temperature in celsius, default 25.

        Example usage::

          my_tmart.set_water(water_salinity=35, water_temperature=20)
          
        '''
        self.water_salinity = water_salinity
        self.water_temperature = water_temperature
        

    # Initiate a wavelength- or band-specific atmosphere, whitecaps properties and water refractive index 
    def _init_atm(self,band): 
        
        if self.sensor_coords is None: # Edit!!!
            print ("WARNING: geometry missing, use set_geometry before you run")
        else:
            
            # Atmospheric profile and aerosol SPF
            self.atm_profile_wl, self.aerosol_SPF_wl = self.Atmosphere._wavelength(self.wl,band)
            
            # Fraction and reflectance of whitecaps 
            self.F_wc_wl, self.R_wc_wl = find_R_wc(wl=self.wl, wind_speed = self.wind_speed)
            
            # Water refractive index 
            self.water_refraIdx_wl = RefraIdx(self.water_salinity,self.water_temperature,self.wl)
            # self.water_refraIdx_wl = 1.34


    # User interface 
    def run(self, wl, band = None, n_photon=10_000, nc='auto', njobs=100, print_on=False, output_flux=False): 
        '''Run with multiple processing 
        
        Arguments:

        * ``wl`` -- wavelength in nm.
        * ``band`` -- overwrite ``wl`` with a 6S band object. We still need to specify ``wl`` because it is used in interpolating spectral SPF.
        * ``n_photon`` -- number of photons to use in MC simulation, default 10,000.
        * ``nc`` -- number of CPU cores to use in multiprocessing, default automatic. 
        * ``njobs`` -- dividing the jobs into n portions in multiprocessing, default 80. 
        
        Return:

        * Movement information of photons.
        
        Example usage::

          n_photon = 100_00
          nc = 10
          njobs = 100
          results = my_tmart.run(wl=wl, n_photon=n_photon,nc= nc,njobs= njobs)
          
        '''
        
        self.wl = wl 
        self.print_on = print_on
        self.plot_on = False # don't even try it 
        self.output_flux = output_flux
        self._init_atm(band)
        
        
        if nc=='auto':
            nc = cpu_count()
        else:
            nc = nc
            
        print("\n========= Initiating T-Mart =========")
        print(f"Number of photons: {n_photon}")
        print(f'Using {nc} core(s)')
        
        n = n_photon
        
        # Split all photon IDs into numpy arrays 
        part_count = [n/njobs for i in range(njobs)]
        part_count = np.array_split(range(n_photon), njobs)
        
        
        print(f"Number of job(s): {njobs}")
        print('Wavelength: ' + str(self.wl) + ' nm')
        print('Aerosol type: ' + str(self.Atmosphere.aerosol_type))
        print('AOT at 550 nm: ' + str(self.Atmosphere.aot550)) 
        print('Photon\'s initial direction: ' + str( np.round(self.target_pt_direction,2) ))
        print('Solar angle: ' + str( np.round(self.sun_dir, 2) ))
        print("=====================================")
        

        pool = ProcessingPool(processes=nc)
        time.sleep(0.5)
        
        # manual print 
        results_temp = pool.amap(self._run,part_count) # Async
        
        if njobs>1:
            _track_job(results_temp)
        
        results = results_temp.get()
        results = np.vstack(results)

        return results 

        
    # Distribute runs to processors     
    def _run(self,part_count):
    
        pts_stat = np.empty([0,13])
        
        for i in part_count:
            
            if self.print_on:
                print("\n---------- Running Photon " + str(i) + " ----------")
            
            pt_stat = self._run_single_photon(i)
            pts_stat = np.vstack([pts_stat, pt_stat])
      
        return pts_stat
    
    
    def run_plot(self, wl, band = None, plot_on=True, plot_range=[0,100_000,0,100_000,0,100_000]): 
        '''Run a single photon and plot, print the details of photon movements. 
        To observe the photon movements, mostly for debugging purposes. 
        
        Arguments:

        * ``wl`` -- wavelength in nm.
        * ``band`` -- overwrite ``wl`` with a 6S band object. We still need to specify ``wl`` because it is used in interpolating spectral SPF.
        * ``plot_on`` -- Boolean, if plot the movements. 
        * ``plot_range`` -- List, [xmin, xmax, ymin, ymax, zmin, zmax]
        
        Return:

        * Movement information of a photon.

        Example usage::

          results = my_tmart.run_plot(wl=wl, plot_on=True, plot_range=[0,100_000,0,100_000,0,100_000])
          
        '''
        
        
        print("\n====== Running and Plotting T-Mart Single Photon ======")
        self.wl = wl 
        self.print_on = True    # Always print the details of photon movements 
        self.plot_on = plot_on  # Default plot, may turn off 
        self.plot_range = plot_range
        self._init_atm(band)
        
        return self._run_single_photon(0)
    
    














# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.



from .Surface import Surface, SpectralSurface
from .Atmosphere import Atmosphere

from .Tmart import Tmart

from .tm_calcref import calc_ref 
from .tm_geometry import dirP_to_coord

from tmart import AEC
from tmart import surface_rho





# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# %matplotlib qt

# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)


# Actual imports 
import tmart
import numpy as np
from Py6S.Params.atmosprofile import AtmosProfile
import Py6S


wl = [400,800,100]

wl = 800


rho = tmart.surface_rho.calculate(wl=wl, viewing_zenith=30, solar_zenith=30.88, 
                                  relative_azimuth=90, aot550=0.05, wind_speed = 8, n_photon=10_000)


print(rho)









# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Correct for the non-linearity of the ratio of environmental irradiance to surface reflectance for homogeneous Lambertian surfaces. 

from Py6S import *

def irradiance_correction(image, wl_RC, band = None,
                          tm_vza = 0, tm_vaa = 0, tm_sza = 0, tm_saa = 0, 
                          atm_profile = None, 
                          aerosol_type = 'Maritime', aot550 = 0):
    '''
    

    Parameters
    ----------
    image : numpy array 
        surface reflectance.
    wl_RC : TYPE
        DESCRIPTION.
    band : TYPE, optional
        DESCRIPTION. The default is None.
    tm_vza : TYPE, optional
        DESCRIPTION. The default is 0.
    tm_vaa : TYPE, optional
        DESCRIPTION. The default is 0.
    tm_sza : TYPE, optional
        DESCRIPTION. The default is 0.
    tm_saa : TYPE, optional
        DESCRIPTION. The default is 0.
    atm_profile : TYPE, optional
        DESCRIPTION. The default is None.
    aerosol_type : TYPE, optional
        DESCRIPTION. The default is 'Maritime'.
    aot550 : TYPE, optional
        DESCRIPTION. The default is 0.

    Returns
    -------
    image_out : TYPE
        DESCRIPTION.

    '''
    
    import numpy as np
    import sys
    
    
    # An unknown bug of 6S unable to handle 551nm
    if wl_RC == 0.551: wl_RC = 0.55
    
    median_image = np.median(image)
    
    # SixS
    s = SixS()
    s.geometry = Geometry.User()
    s.geometry.solar_z = tm_sza
    s.geometry.solar_a = tm_saa
    s.geometry.view_z = tm_vza
    s.geometry.view_a = tm_vaa
    
    # Atmosphere
    if atm_profile is None:
        s.atmos_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer)
    else: 
        s.atmos_profile = AtmosProfile.UserWaterAndOzone(atm_profile['water_vapour']/10, atm_profile['ozone']/1000)

    # Aerosol 
    if aerosol_type == 'Maritime':
        aerosol_profile = s.aero_profile = AeroProfile.PredefinedType(AeroProfile.Maritime)
    elif aerosol_type == 'Continental':
        aerosol_profile = s.aero_profile = AeroProfile.PredefinedType(AeroProfile.Continental)
    elif isinstance(aerosol_type, float):
        r_mar = aerosol_type
        r_con = 1 - r_mar
        s.aero_profile = AeroProfile.User(soot = 0.01*r_con, water = 0.05*r_mar + 0.29*r_con, 
                                          oceanic = 0.95*r_mar, dust = 0.7*r_con)
    else:
        sys.exit('Unrecgnized aerosol type')
        
    s.aot550 = aot550
    
    # Elevation 
    s.altitudes.set_sensor_satellite_level()
    s.altitudes.set_target_custom_altitude(0)
    
    # Wavelength and band 
    if band is None:  
        s.wavelength = Wavelength(wl_RC) 
    else: 
        s.wavelength = band
        
    # Lists of TOA direct reflectance and corresponding surface reflectance 
    list_R_dir = []
    list_SR = [0.25,0.5,0.75,1]
    
    for SR in list_SR:
        s.ground_reflectance = GroundReflectance.HomogeneousLambertian(SR)
        s.run()
        list_R_dir.append(s.outputs.values['pixel_reflectance'])
    
    
    # Calculate the ratio 
    array_R_dir = np.array(list_R_dir)
    array_SR = np.array(list_SR)
    array_ratio = array_R_dir / array_SR
    fit_ratio_SR = np.polyfit(array_SR , array_ratio, 2)
    p = np.poly1d(fit_ratio_SR)
    median_ratio = p(median_image)
    
    # Ratio of all pixels 
    ratio = p(image)
    
    # Relative to median 
    correction = ratio / median_ratio
    
    # Do not correct pixels brighter than mean 
    correction[correction>1] = 1
    image_out = image * correction
    max_correction = correction.min()
    max_correction_percent = str(round((1 - max_correction)*100, 2))
    print('Maximum change in pixel value: ' + max_correction_percent + '%')
    
    return image_out




if __name__ == "__main__":
    import numpy as np
    import Py6S
    
    
    # import netCDF4 as nc4
    # file = 'test.nc'
    # with nc4.Dataset(file, 'r')  as dset:
    #     a_test = dset['rhot_551'][:]
    
    band = Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_08)
    
    a_test = np.array([[0,0.1],[0.5,1]])
    
        
    test = reflectance_correction(a_test, 0.551, aot550=1, aerosol_type=0.999)
    
    print(test)
    
    





# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.




# Extract aerosol type from NASA Ocean Color ancillary data 
def anci_get_AER(files, metadata): 
    
    print('\nRetrieving aerosol type at center of image: ')
    r_maritime_0 = _get_AER(files[0], metadata['lat'], metadata['lon'])
    r_maritime_1 = _get_AER(files[1], metadata['lat'], metadata['lon'])
    
    # Interpolate aerosol linearly in time
    minute = float(metadata['time'][14:16])
    second = float(metadata['time'][17:19])
    ratio = minute/60 + second/3600
    r_maritime = r_maritime_1 * ratio + r_maritime_0 * (1-ratio)
    print('Interpolated ratio of maritime aerosol in maritime/continental mixture: ' + str(r_maritime[0]))
    print('Angstrom exponent: {:.2f}'.format(r_maritime[1]))
    print('Single scattering albedo: {:.2f}'.format(r_maritime[2]))
    return r_maritime
    
    
# Find value of ancillary data at lat and lon
def _get_AER(file, lat, lon):
    
    import netCDF4 as nc4
    import numpy as np
    from scipy import interpolate
    
    # from https://stackoverflow.com/questions/31820107/is-there-a-numpy-function-that-allows-you-to-specify-start-step-and-number
    def linspace(start, stop, step=1.):
        import numpy as np
        """
        Like np.linspace but uses step instead of num
        This is inclusive to stop, so if start=1, stop=3, step=0.5
        Output is: array([1., 1.5, 2., 2.5, 3.])
        """
        return np.linspace(start, stop, int((stop - start) / step + 1))
    
    # return ratio of maritime, rest being continental 
    def categorize_aer(i,maritime,continental):
        if maritime < continental: 
            if i < maritime: return 1
            elif i > continental: return 0
            else: return abs(i - maritime) / abs(continental - maritime)
        else:
            if i > maritime: return 1
            elif i < continental: return 0
            else: return abs(maritime - i) / abs(maritime - continental)
    
    nc = nc4.Dataset(file, 'r') 
    
    lons = linspace(nc.getncattr('geospatial_lon_min'),
                    nc.getncattr('geospatial_lon_max'),
                    nc.getncattr('geospatial_lon_resolution'))
    
    lats = linspace(nc.getncattr('geospatial_lat_min'),
                    nc.getncattr('geospatial_lat_max'),
                    nc.getncattr('geospatial_lat_resolution'))
    
    aer_ANG = nc['TOTANGSTR'][:]
    aer_SSA = nc['TOTSCATAU'][:] / nc['TOTEXTTAU'][:]
    aer_AOT = nc['TOTSCATAU'][:]
    
    interp_ANG = interpolate.RegularGridInterpolator((lats,lons), aer_ANG)
    i_ANG = interp_ANG([lat,lon])[0]
    
    interp_SSA = interpolate.RegularGridInterpolator((lats,lons), aer_SSA)
    i_SSA = interp_SSA([lat,lon])[0]
    
    interp_AOT = interpolate.RegularGridInterpolator((lats,lons), aer_AOT)
    i_AOT = interp_AOT([lat,lon])[0]
    
    # print('Angstrom exponent: ' + str(i_ANG))
    # print('SSA: ' + str(i_SSA))
    
    # according the the two 
    maritime_ANG = categorize_aer(i_ANG, maritime=0.265, continental=1.132)
    maritime_SSA = categorize_aer(i_SSA, maritime=0.98903, continental=0.89319)
    ratio_maritime = (maritime_ANG + maritime_SSA) / 2
    
    # print('ratio_maritime: ' + str(ratio_maritime))
    
    return np.array([ratio_maritime, i_ANG, i_SSA, i_AOT])





# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# %matplotlib qt


# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)

import tmart
import numpy as np
from Py6S.Params.atmosprofile import AtmosProfile

# Specify wavelength in nm
wl = 800

# Distance to shore in m 
dist = 1000

# Built-in spectral libraries 
water = tmart.SpectralSurface('water_chl1')
vegetation = tmart.SpectralSurface('vegetation')


### DEM and reflectance ###

# Three same-size numpy arrays are needed
image_DEM = np.full((2, 2), 0) # in meters
image_reflectance = np.array([[vegetation.wl(wl),vegetation.wl(wl)],
                              [water.wl(wl),water.wl(wl)]]) 
image_isWater = np.array([[0,0],[1,1]])

# pixel width and length, in meters 
cell_size = 20_000

# Synthesize a surface object
my_surface = tmart.Surface(image_DEM,image_reflectance,image_isWater,cell_size)  
# Set background information, 1 or 2 background surfaces can be set;
# If 2: the first background is the one closer to [0,0]
my_surface.set_background(bg_ref        = [vegetation.wl(wl),water.wl(wl)], 
                          bg_isWater    = [0,1], # water
                          bg_elevation  = 0, # elevation of both background
                          bg_coords     = [[0,20_000],[40_000,20_000]]) # a line dividing two background                                    

### Atmosphere ###

# Atmophere profile comes from 6S
atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 
aerosol_type = 'Maritime' 
aot550 = 0.1

# Synthesize an atmosphere object    
my_atm = tmart.Atmosphere(atm_profile, aot550, aerosol_type)   

### Running T-Mart ###

# Make a T-Mart object 
my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm, shadow=False)
my_tmart.set_wind(wind_speed=10, wind_azi_avg = True)
my_tmart.set_water(water_salinity=35, water_temperature=20)

# Specify the sensor's position (x, y, z), viewing direction relative 
# to the sensor (zenith, azimuth), sun's direction relative to the target 
# (zenith, azimuth)
my_tmart.set_geometry(target_coords=[30000,20000+dist], 
                      target_pt_direction=[150,0],
                      sun_dir=[30,90])  

# Run single photon and plot 
# results = my_tmart.run_plot(wl=wl, plot_on=True, plot_range=[0,100_000,0,100_000,0,100_000])

if __name__ == "__main__":
    
    n_photon = 100_000
    results = my_tmart.run(wl=wl, n_photon=n_photon)
    
    # Calculate reflectances using recorded photon information 
    R = tmart.calc_ref(results, detail=True)
    
    for k, v in R.items():
        print(k, '     ' , v)
    
    
    
    
    


# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# %matplotlib qt


# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)


import tmart
import numpy as np
from Py6S.Params.atmosprofile import AtmosProfile
import Py6S

# Specify wavelength in nm
band = Py6S.Wavelength(Py6S.PredefinedWavelengths.S2A_MSI_08)
wl = 833

### DEM and reflectance ###
image_DEM = np.array([[0,0],[0,0]]) # in meters
image_reflectance = np.array([[0.00,0.00],[0.00,0.00]]) # unitless     
# image_isWater = np.zeros(image_DEM.shape) # 1 is water, 0 is land
image_isWater = np.array([[1,1],[1,1]]) 


# Synthesize a surface object
my_surface = tmart.Surface(DEM = image_DEM,
                           reflectance = image_reflectance,
                           isWater = image_isWater,
                           cell_size = 10_000)  
           
my_surface.set_background(bg_ref        = 0.0, # background reflectance
                          bg_isWater    = 1, # if is water
                          bg_elevation  = 0, # elevation of both background
                          bg_coords     = [[0,0],[10,10]]) # a line dividing the two background                                    
                    
### Atmosphere ###
atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 


my_atm = tmart.Atmosphere(atm_profile, aot550 = 0.11275386685706923, aerosol_type = 0.287 )

### Running T-Mart ###
my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm, shadow=False)
my_tmart.set_wind(wind_speed=1, wind_azi_avg = True)

sensor_coords=[51,50,130_000]


my_tmart.set_geometry(sensor_coords=sensor_coords, 
                      target_pt_direction=[170.52804413432926, 191.91873559828522],
                      sun_dir=[30.9608405674786, 323.9885587375248])

# my_tmart.set_geometry(sensor_coords=sensor_coords, 
#                       target_pt_direction=[180-30, 0],
#                       sun_dir=[10, 0])


results = my_tmart.run(wl=wl, band=band, n_photon=10_000)

# Calculate reflectances using recorded photon information 
R = tmart.calc_ref(results,detail=True)
for k, v in R.items():
    print(k, '     ' , v)



my_atm_profile = my_tmart.atm_profile_wl
np.sum(my_atm_profile.ot_mie)














# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Identify satellite sensor, currently only supports S2 and L8/L9 

def identify_sensor(file):
    import os
    
    strings = file.split(".")
    base_name = os.path.basename(file)
    
    # remove AEC_ if exists 
    if base_name[0:4] == 'AEC_':
        base_name = base_name[4:]
    
    if strings[-1] == 'SAFE':
        if   base_name[0:3] == 'S2A': sensor = 'S2A'
        elif base_name[0:3] == 'S2B': sensor = 'S2B'
    else:
        if   base_name[0:4] == 'LC08': sensor = 'L8'
        elif base_name[0:4] == 'LC09': sensor = 'L9'

    print('\nIdentified sensor: ' + str(sensor) +  '\n')
    return sensor
    
    
    


# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Overall control of AEC

def run(file, username, password, overwrite=False, AOT = None, n_photon = 100_000):
    '''Run adjacency-effect correction on satellite files. Currently only supports Sentinel-2 MSI and Landsat 8 OLI products. See 'Introduction - Adjacency-Effect Correction' for detailed instructions.
    
    Arguments:
        
    * ``file`` -- String. Path to satellite files. 
    * ``username`` -- String. Username of EarthData account.
    * ``password`` -- String. Password of EarthData account.
    * ``overwrite`` -- Boolean. If overwrite the existing files. The default is False and it creates a folder in the same directory that starts with AEC in the name
    * ``AOT`` -- Float. AOT at 550nm, calculated in T-Mart if not specified. 'MERRA2' ancillary data is another option. 
    * ``n_photon`` -- Int. Number of photons in each T-Mart run, 100_000 is recommended for accurate results.   

    Example usage::
        
        file = 'user/test/S2A_MSIL1C_20160812T143752_N0204_R096_T20MKB_20160812T143749.SAFE'
        username = 'abcdef'
        password = '123456'
        tmart.AEC.run(file, username, password)

    '''
    
    import tmart
    import sys, os
    import time

    # If not overwrite, creates a new folder in the same directory  
    if not overwrite:
        from distutils.dir_util import copy_tree
        file_parts = file.rsplit('/',1)
        file_new = file_parts[0] + '/AEC_' + file_parts[1]
        if os.path.exists(file_new):
            print('\nNew directory ' + str(file_new) + ' already exists, proceeds to AEC. \n')
        else: 
            print('\nCopying from ' + str(file) + ' to ' + str(file_new) + '... \n')
            copy_tree(file, file_new)
        file = file_new

    # Check if AEC has been done, if so, exit 
    AEC_record = '{}/{}'.format(file,'AEC_completed.txt')
    if os.path.exists(AEC_record): 
        sys.exit('Record of adjacency correction found: ' + str(AEC_record) + 
                 '. Corrcting for the AE twice is not recommended and will lead to unpredictable results. ' + 
                 'If you’re confident that the imagery has not been corrected for the AE, delete this file to proceed.')

    # Start logging in txt file
    orig_stdout = sys.stdout
    log_file = 'tmart_log_' + time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime(time.time())) + '.txt'
    log_file = '{}/{}'.format(file,log_file)

    class Logger:
        def __init__(self, filename):
            self.console = sys.stdout
            self.file = open(filename, 'w')
            self.file.flush()
        def write(self, message):
            self.console.write(message)
            self.file.write(message)
        def flush(self):
            self.console.flush()
            self.file.flush()

    sys.stdout = Logger(log_file)
    print('System time: ' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time())))
    print('File: ' + str(file))
    
    # Read configuration
    config = tmart.AEC.read_config()

    # Identify sensor
    print('\nReading image files: ')
    print(file)
    sensor = tmart.AEC.identify_sensor(file)
    
    # Extract metadata
    if sensor == 'S2A' or sensor == 'S2B':
        metadata = tmart.AEC.read_metadata_S2(file,config, sensor)
    elif sensor == 'L8':
        metadata = tmart.AEC.read_metadata_L8(file,config)
    else: sys.exit('Warning: unrecognized sensor')
    metadata['sensor'] = sensor
    
    # Print metadata 
    print('Metadata: ')
    for k, v in metadata.items():
        print(str(k) + ': '  + str(v))
        
    # Get ancillary information from NASA Ocean Color
    anci = tmart.AEC.get_ancillary(metadata, username, password)
    
    # Compute cloud and non-Water masks 
    print('\nComputing masks: ')
    mask_cloud = tmart.AEC.compute_masks(metadata, config, 'cloud')
    mask_all   = tmart.AEC.compute_masks(metadata, config, 'all')   
    print('Done')
    
    # Estimate AOT using the 860nm band 
    if AOT == None:
        print('\nEstimating AOT from the NIR band: ')
        AOT = tmart.AEC.get_AOT(metadata, config, anci, mask_cloud, mask_all, n_photon)
    elif AOT == 'MERRA2':
        print('\nUsing AOT from MERRA2: ' + str(AOT))
        AOT = anci['AOT_MERRA2']
    else:
        print('\nUser input AOT: ' + str(AOT))
    
    # Write atm information
    tmart.AEC.write_atm_info(file, anci, AOT)
    print('\nWrote aerosol and atmosphere information.')
    
    # Make a record file for AEC 
    file_AEC_record = open(AEC_record,"w")
    file_AEC_record.flush()
    
    # AEC for each of the specified bands 
    for i in range(len(metadata['AEC_bands_name'])):
        AEC_band_name = metadata['AEC_bands_name'][i] 
        AEC_band_6S = metadata['AEC_bands_6S'][i]
        wl = metadata['AEC_bands_wl'][i]
        print('\n============= AEC: {} ==================='.format(AEC_band_name))
        tmart.AEC.AEC(AEC_band_name, AEC_band_6S, wl, AOT, metadata, config, anci, mask_cloud, mask_all, n_photon)
        file_AEC_record.write(str(AEC_band_name) + '\n')
        file_AEC_record.flush()
        
    # Close AEC record 
    file_AEC_record.close()
    
    # Stop logging 
    sys.stdout = orig_stdout

    return 0
    
    


# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tmart",                     # This is the name of the package
    version="2.0.7",                       
    author="Yulun Wu",                     # Full name of the author
    description="Modelling and correcting for the adjacency effect in aquatic remote sensing",
    long_description=long_description,      # Long description read from the the readme file
    long_description_content_type="text/markdown",
    # packages=setuptools.find_packages(),    # List of all python modules to be installed
    packages = ['tmart','tmart.surface_rho','tmart.AEC','tmart.ancillary','tmart.ancillary.aerosolSPF','tmart.config'],
    include_package_data=True,
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        "Operating System :: OS Independent",
    ],                                      # Information to filter the project on PyPi website
    python_requires='>=3.8',                # Minimum version requirement of the package
    # py_modules=["tmart"],             # Name of the python package
    # package_dir={'':'tmart/'},     # Directory of the source code of the package
    license_files=('license.txt'),
    install_requires=['Py6S','scipy','pathos','matplotlib','netCDF4','rasterio','mgrs','geopandas','requests']  # Install other dependencies if any
)


# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

def calculate(wl, viewing_zenith, solar_zenith, relative_azimuth, 
              aot550=0.05, wind_speed=3,n_photon = 100_000,
              atm_profile=None,aerosol_type=None,spectral_surface=None,as_pandas_df=True):
    
    '''Calculate sea-surface reflectance factor. See Wu et al. 2023, Lee et al. 2010, or Mobley 1999 for full description. 

    Arguments:
    
    * ``wl`` -- central wavelength in nm, or a list for iteration: [starting_wavelength, ending wavelength, interval].
    * ``viewing_zenith`` -- viewing zenith angle in degrees.
    * ``solar_zenith`` -- solar zenith angle in degrees.
    * ``relative_azimuth`` -- relative azimuth angle in degrees.
    * ``aot550`` -- aerosol optical thickness at 550nm.
    * ``wind_speed`` -- wind speed in m/s.
    * ``n_photon`` -- number of photons in each T-Mart run. 100_000 is recommended, 10_000 can be used to view quick results. 
    * ``atm_profile`` -- AtmosProfile object from Py6S. Default 'MidlatitudeSummer'
    * ``aerosol_type`` -- 'BiomassBurning', 'Continental', 'Desert', 'Maritime', 'Stratospheric' or 'Urban', as provided by 6S. Default 'Maritime'
    * ``spectral_surface`` -- a SpectralSurface object in T-Mart. Default 'water_chl1'
    * ``as_pandas_df`` -- If true, return a pandas dataframe, else return a dictionary
    
    Example usage:: 
    
      tmart.surface_r.calculate(wl=800, viewing_zenith=40, solar_zenith=30, relative_azimuth=135)

    '''
    
    import tmart
    import numpy as np
    import pandas as pd
    from Py6S.Params.atmosprofile import AtmosProfile
    
    ### Input 
    
    if atm_profile==None: 
        atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 
    
    if aerosol_type==None: 
        aerosol_type = 'Maritime' 
    
    if spectral_surface == None:
        spectral_surface = tmart.SpectralSurface('water_chl1')
    
    
    rho_output = []
    
    
    
    if isinstance(wl, list):
        wls = range(wl[0],wl[1]+wl[2],wl[2])
    else:
        wls = [wl]
    
    
    for wl in wls:
    
        ### L_sky ###
        
        # Three same-size numpy arrays are needed
        image_DEM = np.array([[0,0],[0,0]]) # in meters
       
        
        image_reflectance = np.array([[0,0],
                                      [0,0]]) # unitless     
        
        image_isWater = np.array([[1,1],[1,1]])# 1 is water, 0 is land
        
        # pixel width and length, in meters 
        cell_size = 0.1
        
        # Synthesize a surface object
        my_surface = tmart.Surface(image_DEM,image_reflectance,image_isWater,cell_size)  
        # Set background information, 1 or 2 background surfaces can be set;
        # If 2: the first background is the one closer to [0,0]
        my_surface.set_background(bg_ref        = spectral_surface.wl(wl), 
                                  bg_isWater    = 1,
                                  bg_elevation  = 0, # elevation of both background
                                  bg_coords     = [[0,0],[10,10]]) # a line dividing two background
        
        # Synthesize an atmosphere object     
        my_atm = tmart.Atmosphere(atm_profile, aot550, aerosol_type)
        
        ### Running T-Mart ###
        my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm)
        my_tmart.set_wind(wind_speed=wind_speed,wind_azi_avg=True)
        my_tmart.set_water(water_salinity=35, water_temperature=20)
        my_tmart.set_geometry(sensor_coords=[0.075,0.05,0.001], 
                              target_pt_direction=[viewing_zenith,relative_azimuth],
                              sun_dir=[solar_zenith,0])
        
        results = my_tmart.run(wl=wl, n_photon=n_photon)
        R1 = tmart.calc_ref(results, n_photon=n_photon)
        R1['Wavelength'] = wl
        R1['Type'] = 'L_sky'
        for k, v in R1.items():
    	    print(k, '     ' , v)
        
        ### L_sr ###
        
        my_tmart.set_geometry(sensor_coords=[0.075,0.05,0.001], 
                              target_pt_direction=[180-viewing_zenith,relative_azimuth],
                              sun_dir=[solar_zenith,0])
        
        
        results = my_tmart.run(wl=wl, n_photon=n_photon)
        R2 = tmart.calc_ref(results, n_photon=n_photon)
        R2['Wavelength'] = wl
        R2['Type'] = 'L_sr'
        for k, v in R2.items():
    	    print(k, '     ' , v)
        
        
        ### calculate rho ###
        
        rho = R2['R_total'] / R1['R_total']
        
        print('Sea-surface reflectance factor: ' + str(rho))
        
        rho_output.append({'wavelength': wl, 'rho': rho})
    

    if as_pandas_df: rho_output = pd.DataFrame(rho_output)
        
    
    return rho_output 







# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Read L8/9 metadata

def read_metadata_L8(file, config):
    import pandas as pd
    import glob, os, Py6S
    import math 
    
    # MTL file 
    mtl = glob.glob('{}/{}'.format(file, '*MTL.txt'))[0]
    df = pd.read_csv(mtl, delimiter = '=', skipinitialspace=True)
    
    # Get the first and second columns by integer positions
    first_column = df.iloc[:, 0]
    first_column = first_column.str.replace(' ', '')
    second_column = df.iloc[:, 1]

    # Convert to dictionary
    result_dict = dict(zip(first_column, second_column))

    # Extract values 
    SUN_AZIMUTH = float(result_dict['SUN_AZIMUTH'])
    SUN_ELEVATION = float(result_dict['SUN_ELEVATION'])
    time = result_dict['DATE_ACQUIRED']+'T'+result_dict['SCENE_CENTER_TIME']
    lat = (float(result_dict['CORNER_UL_LAT_PRODUCT']) + float(result_dict['CORNER_UR_LAT_PRODUCT']) + 
           float(result_dict['CORNER_LL_LAT_PRODUCT']) + float(result_dict['CORNER_LR_LAT_PRODUCT']))/4
    lon = (float(result_dict['CORNER_UL_LON_PRODUCT']) + float(result_dict['CORNER_UR_LON_PRODUCT']) + 
           float(result_dict['CORNER_LL_LON_PRODUCT']) + float(result_dict['CORNER_LR_LON_PRODUCT']))/4
    tm_sun_dir=[90.0-SUN_ELEVATION, (SUN_AZIMUTH+270)%360] # mean_saa = 0 => sun is in the north => 270 in T-Mart
    
    # Create dictionary 
    metadata =  {'file': file,
                 'granule': file,
                 'lat': lat,
                 'lon': lon,
                 'time':time,
                 'saa':SUN_AZIMUTH,
                 'sza':90.0-SUN_ELEVATION,
                 'vaa':0,
                 'vza':0,
                 'tm_pt_dir': [180,0],
                 'tm_sun_dir': tm_sun_dir}
    
    # Scaling factors 
    metadata['B1_mult'] = float(result_dict['REFLECTANCE_MULT_BAND_1']) 
    metadata['B2_mult'] = float(result_dict['REFLECTANCE_MULT_BAND_2']) 
    metadata['B3_mult'] = float(result_dict['REFLECTANCE_MULT_BAND_3']) 
    metadata['B4_mult'] = float(result_dict['REFLECTANCE_MULT_BAND_4']) 
    metadata['B5_mult'] = float(result_dict['REFLECTANCE_MULT_BAND_5']) 
    metadata['B6_mult'] = float(result_dict['REFLECTANCE_MULT_BAND_6']) 
    metadata['B7_mult'] = float(result_dict['REFLECTANCE_MULT_BAND_7']) 
    metadata['B9_mult'] = float(result_dict['REFLECTANCE_MULT_BAND_9']) 
    metadata['B1_add'] = float(result_dict['REFLECTANCE_ADD_BAND_1']) 
    metadata['B2_add'] = float(result_dict['REFLECTANCE_ADD_BAND_2']) 
    metadata['B3_add'] = float(result_dict['REFLECTANCE_ADD_BAND_3']) 
    metadata['B4_add'] = float(result_dict['REFLECTANCE_ADD_BAND_4']) 
    metadata['B5_add'] = float(result_dict['REFLECTANCE_ADD_BAND_5']) 
    metadata['B6_add'] = float(result_dict['REFLECTANCE_ADD_BAND_6']) 
    metadata['B7_add'] = float(result_dict['REFLECTANCE_ADD_BAND_7'])
    metadata['B9_add'] = float(result_dict['REFLECTANCE_ADD_BAND_9'])
    
    # Band and mask files 
    files = os.listdir(file)
    for image in files:
        if image[0]=='.':continue
        tmp = image.split('_')
        if tmp[-1][0] == 'B' and tmp[-1][-3:] == 'TIF':
            metadata[tmp[-1].split('.')[0]]  = os.path.join(file,image)
        elif tmp[-1] == 'PIXEL.TIF': 
            metadata['cloud_mask'] = os.path.join(file,image)
        else:
            pass
    
    # Resolution for masks 
    metadata['mask_res'] = [30, int( 30 *  int(config['reshape_factor_L8']))]
    
    # High TOA bands for masks 
    highTOA_band_names = ['B1','B2','B3','B4','B5','B6','B7'] # bands to use in the highTOA mask 
    highTOA_band_names.remove(config['L8_SWIR_band'])
    metadata['highTOA_band_names'] = highTOA_band_names

    # Masks 
    metadata['cirrus_mask'] = config['L8_cirrus_band']
    metadata['SWIR_mask'] = config['L8_SWIR_band']
  
    # Bands to be AECed
    metadata['AEC_bands_name'] = ['B1','B2','B3','B4','B5','B6','B7']
    metadata['AEC_bands_6S'] = [Py6S.Wavelength(Py6S.PredefinedWavelengths.LANDSAT_OLI_B1),
                                Py6S.Wavelength(Py6S.PredefinedWavelengths.LANDSAT_OLI_B2),
                                Py6S.Wavelength(Py6S.PredefinedWavelengths.LANDSAT_OLI_B3),
                                Py6S.Wavelength(Py6S.PredefinedWavelengths.LANDSAT_OLI_B4),
                                Py6S.Wavelength(Py6S.PredefinedWavelengths.LANDSAT_OLI_B5),
                                Py6S.Wavelength(Py6S.PredefinedWavelengths.LANDSAT_OLI_B6),
                                Py6S.Wavelength(Py6S.PredefinedWavelengths.LANDSAT_OLI_B7)]
    
    # Center wavelength, source: http://gsics.atmos.umd.edu/pub/Development/20171106/5k_Ong_Landsat8_Lunar_Calibrations.pdf
    metadata['AEC_bands_wl'] = [443, 482, 561.4, 654.6, 864.7, 1608.9, 2200.7]
    
    # Others 
    metadata['resolution'] = 30
    metadata['reshape_factor'] = int(config['reshape_factor_L8'])
    metadata['window_size'] = int(config['window_size'])
    metadata['AEC_resolution'] = int(config['reshape_factor_L8']) * 30 # resolution 
    metadata['height'] = int(result_dict['REFLECTIVE_LINES']) 
    metadata['width'] = int(result_dict['REFLECTIVE_SAMPLES']) 
    
    # Height and width for AEC, with a few extra rows and columns 
    metadata['AEC_height'] = math.ceil(metadata['height'] / int(config['reshape_factor_L8'])) * int(config['reshape_factor_L8'])
    metadata['AEC_width'] = math.ceil(metadata['width'] / int(config['reshape_factor_L8'])) * int(config['reshape_factor_L8'])

    return metadata
    
    
    
    
 


# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Read configuration file 

def read_config():
    import os
    import pandas as pd
    
    # Directory 
    one_up = os.path.dirname(os.path.dirname(__file__ ))
    file = os.path.join(one_up, 'config/config.txt')

    # Read file 
    df = pd.read_csv(file, skiprows=0, skipinitialspace=True,on_bad_lines='skip')

    # Remvoe comments 
    df = df.drop(df[df.iloc[:,0].str.startswith('#')].index)
    
    # Extract info 
    df[['left', 'right']] = df.iloc[:,0].str.replace(' ', '').str.split('=', expand=True)
    result_dict = dict(zip(df['left'], df['right']))
    print('\nT-Mart configuration (editable at {}): '.format(file))
    for k, v in result_dict.items():
        print(str(k) + ': '  + str(v))    
    
    return result_dict
    
    
    



# This file is part of T-Mart.
#
# Copyright 2023 Yulun Wu.
#
# T-Mart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.




# Analyze the output of TMart and differentiate direct, env and atm intrinsic reflectances  

import numpy as np 
import pandas as pd
import sys
from copy import copy




def calc_ref(df, n_photon = None, detail = False):
    '''Analyze the results of T-Mart and calculate reflectances. 
    
    Arguments:

    * ``df`` -- Results from T-Mart runs
    * ``n_photon`` -- Specify the number of photons in the run when firing the photon upwards. If not specified, the number of unique pt_id will be used. This can lead to errors when photons were fired upwards because some photons will not have pt_id.
    * ``detail`` -- differentiate coxmunk, whitecap, water-leaving and land contributions

    Output:

    * A list of atmospheric intrinsic reflectance, direct reflectance, environmental reflectance and total reflectance. 

    Example usage::

      R = tmart.calc_ref(results)
      for k, v in R.items():
          print(k, '\t ' , v)

    '''
    
    print('=====================================')
    print('Calculating radiometric quantities...')
    
    # Columes: 0 pt_id, 1 movement, 2 L_cox-munk, 3 L_whitecap, 4 L_water, 5 L_land, 
    # 6 L_rayleigh, 7 L_mie, 8 9 10 surface xyz, 11 shadowed, 12 if_env
    
    
    
    
    if n_photon == None:
        n_photon = np.unique(df[:,0]).shape[0]
    
    
    # Keep only the non-shaded ones 
    # df = df[df[:,11] == 0,:]
    
    # df[:,0]
    # df[:,6:8]
    # df[:,12] == 1
    # df[df[:,12] == 1 ,2:6] # if_env and all surface reflectance 
    
    
    R_atm = np.sum(df[:,6:8]) / n_photon
    R_dir = np.sum(df[df[:,12] == 0 ,2:6]) / n_photon # if_env == 1 and all surface reflectance 
    R_env = np.sum(df[df[:,12] == 1 ,2:6]) / n_photon # if_env == 0 and all surface reflectance 
    R_total = R_atm + R_dir + R_env
    
    
    if detail:

    
        R_dir_coxmunk   = np.sum(df[df[:,12] == 0 ,2]) / n_photon
        R_dir_whitecap  = np.sum(df[df[:,12] == 0 ,3]) / n_photon
        R_dir_water     = np.sum(df[df[:,12] == 0 ,4]) / n_photon
        R_dir_land      = np.sum(df[df[:,12] == 0 ,5]) / n_photon
        
        R_env_coxmunk   = np.sum(df[df[:,12] == 1 ,2]) / n_photon
        R_env_whitecap  = np.sum(df[df[:,12] == 1 ,3]) / n_photon
        R_env_water     = np.sum(df[df[:,12] == 1 ,4]) / n_photon
        R_env_land      = np.sum(df[df[:,12] == 1 ,5]) / n_photon        
        
        R_output = {'R_atm':R_atm,
                    'R_dir':R_dir,
                    '_R_dir_coxmunk':R_dir_coxmunk,
                    '_R_dir_whitecap':R_dir_whitecap,
                    '_R_dir_water':R_dir_water,
                    '_R_dir_land':R_dir_land,
                    'R_env':R_env,
                    '_R_env_coxmunk':R_env_coxmunk,
                    '_R_env_whitecap':R_env_whitecap,
                    '_R_env_water':R_env_water,
                    '_R_env_land':R_env_land,
                    'R_total':R_total}

    
    else:      
        R_output = {'R_atm':R_atm,
                    'R_dir':R_dir,
                    'R_env':R_env,
                    'R_total':R_total}
    
    
    return R_output
    


# Outdated 

# def calc_ref2(df, n_photon = None, detail = False, total_only = False):
#     '''Analyze the results of T-Mart and calculate reflectances. 
    
#     Arguments:

#     * ``df`` -- Results from T-Mart runs
#     * ``n_photon`` -- Specify the number of photons in the run when firing the photon upwards. If not specified, the number of unique pt_id will be used. This can lead to errors when photons were fired upwards because some photons will not have pt_id.
#     * ``detail`` -- differentiate coxmunk, whitecap, water-leaving and land contributions
#     * ``total_only`` -- only return total TOA reflectance, this is much faster but gives less details. 
    
#     Output:

#     * A list of atmospheric intrinsic reflectance, direct reflectance, environmental reflectance and total reflectance. 

#     Example usage::

#       R = tmart.calc_ref(results)
#       for k, v in R.items():
#           print(k, '\t ' , v)

#     '''
    
#     print('=====================================')
#     print('Calculating radiative properties...')
    
#     dfpd = pd.DataFrame(data=df, columns = ['pt_id', 'movement', 'type_collision', 'L_coxmunk', 
#                                             'L_whitecap', 'L_water', 'L_land', 'L_rayleigh', 'L_mie','x','y','z','shadowed'])
       
#     # reordered 
    
    
#     dfpd.pt_id =        pd.to_numeric( dfpd.pt_id)
#     dfpd.movement =     pd.to_numeric( dfpd.movement)
#     dfpd.L_coxmunk =    pd.to_numeric( dfpd.L_coxmunk)
#     dfpd.L_whitecap =   pd.to_numeric( dfpd.L_whitecap)
#     dfpd.L_water =      pd.to_numeric( dfpd.L_water)
#     dfpd.L_land =       pd.to_numeric( dfpd.L_land)
#     dfpd.L_rayleigh =   pd.to_numeric( dfpd.L_rayleigh)
#     dfpd.L_mie =        pd.to_numeric( dfpd.L_mie)
#     dfpd.x =            pd.to_numeric( dfpd.x)
#     dfpd.y =            pd.to_numeric( dfpd.y)
#     dfpd.z =            pd.to_numeric( dfpd.z)
#     dfpd.shadowed =     pd.to_numeric( dfpd.shadowed)
    
#     dfpd['env'] = 0 # mark environment reflectance as 1 
    
#     unique_pt_id = (np.unique(dfpd.pt_id))
    
#     if n_photon == None:
#         n_photon = unique_pt_id.shape[0]
    
    
#     ### Total reflectance only 
    
#     if total_only:
#         R_total = np.sum(dfpd.L_coxmunk) + np.sum(dfpd.L_whitecap) + np.sum(dfpd.L_water) + np.sum(dfpd.L_land) + np.sum(dfpd.L_rayleigh) + np.sum(dfpd.L_mie) 
#         R_total = R_total / n_photon
#         return R_total
    

    
#     ### Differentiate atm, dir and env reflectances 
    
#     # Number of rows for each pt_id
#     pt_id_counts = dfpd.pt_id.value_counts()   
    
#     # Isolate those that only have one row to speed up calculation 
#     pt_ids_1 = pt_id_counts.index[pt_id_counts == 1].tolist()
    
#     # Main output data frame 
#     df_output = dfpd[dfpd['pt_id'].isin(pt_ids_1)] # append to this 
    
#     # Extract the ones with more than one pt_id, meaning multiple scattering events 
#     pt_ids_2 = pt_id_counts.index[pt_id_counts > 1].tolist() # pt_ids more than 1 rows 
    
#     '''testing 
    
#     pt_id = 1
    
#     '''
    
#     # Loop through pt_ids, differentiate reflectances 
#     for pt_id in pt_ids_2:
    
#         # Data frame with this pt_id 
#         pt = dfpd[dfpd.pt_id  == pt_id]
        
#         # Each movement of the photon 
#         moves = pt.movement 
        
#         if not moves.is_monotonic_increasing: # check if sorted 
#             sys.exit('pt movement has to be sorted')
        
#         # Find the move with W, L or Ws collision, then add all after to them and stop the loop
#         for move in moves:
            
#             # Entire row 
#             pt_movement = copy(pt[pt.movement == move]) 
            
#             # Type of collision 
#             t_c = pt_movement.type_collision.values[0] 
            
#             if t_c=='W' or t_c=='L':
#                 # Adding all after to L_whitecap, L_water, L_land
                
#                 # If surface is black, quit loop to speed up calculation 
#                 if sum( pt.iloc[move,4:7]) == 0: 
#                     break
                
#                 # Identify all movements after that contribute to this 
#                 pt_mov_after = pt.iloc[move+1:,:]
#                 pt_mov_after_nonShadow = pt_mov_after[pt_mov_after.shadowed==0]
#                 sum_after = np.sum(pt_mov_after_nonShadow.iloc[:,3:9].values)
                
#                 # Total of the single row
#                 total = pt_movement.L_whitecap.values[0] + pt_movement.L_water.values[0] + pt_movement.L_land.values[0]
                
#                 # Ratios 
#                 r_wc = pt_movement.L_whitecap.values[0] / total 
#                 r_water = pt_movement.L_water.values[0] / total
#                 r_land = pt_movement.L_land.values[0] / total 
                
#                 # We calculate ratio before setting shadowed 'total' to 0
#                 if pt_movement.shadowed.values[0] == 1: total = 0
                
#                 total_new = total + sum_after
            
#                 pt_movement.L_whitecap = total_new * r_wc
#                 pt_movement.L_water = total_new * r_water
#                 pt_movement.L_land = total_new * r_land
                
#                 # move>0 means at least one atmospheric scattering happened
#                 if move>0: pt_movement.env = 1
                
#                 # Add to the main data frame 
#                 df_output = pd.concat([df_output,pt_movement])
                
#                 break 
                
#             if t_c=='Ws':    
#                 # Adding all after to L_coxmunk
                
#                 pt_mov_after = pt.iloc[move+1:,:]
#                 pt_mov_after_nonShadow = pt_mov_after[pt_mov_after.shadowed==0]
#                 sum_after = np.sum(pt_mov_after_nonShadow.iloc[:,3:9].values)          
                
#                 if pt_movement.shadowed.values[0] == 1: 
#                     total = 0
#                 else:
#                     total = pt_movement.L_coxmunk.values[0]
                
#                 pt_movement.L_coxmunk = total + sum_after
                
#                 if move>0: pt_movement.env = 1
                
#                 df_output = pd.concat([df_output,pt_movement])
                
#                 break 
            
#             # If rayleigh or mie 
#             df_output = pd.concat([df_output,pt_movement])
    
    
#     R_atm = np.sum(df_output.iloc[:,7:9].values) / n_photon
#     R_dir = np.sum(df_output[df_output.env==0].iloc[:,3:7].values) / n_photon
#     R_env = np.sum(df_output[df_output.env==1].iloc[:,3:7].values) / n_photon
#     R_total = R_atm + R_dir + R_env
    
    
#     if detail:
    
#         R_dir_coxmunk = np.sum(df_output[df_output.env==0].iloc[:,3].values) / n_photon
#         R_dir_whitecap = np.sum(df_output[df_output.env==0].iloc[:,4].values) / n_photon
#         R_dir_water = np.sum(df_output[df_output.env==0].iloc[:,5].values) / n_photon
#         R_dir_land = np.sum(df_output[df_output.env==0].iloc[:,6].values) / n_photon
        
#         R_env_coxmunk = np.sum(df_output[df_output.env==1].iloc[:,3].values) / n_photon
#         R_env_whitecap = np.sum(df_output[df_output.env==1].iloc[:,4].values) / n_photon
#         R_env_water = np.sum(df_output[df_output.env==1].iloc[:,5].values) / n_photon
#         R_env_land = np.sum(df_output[df_output.env==1].iloc[:,6].values) / n_photon        
        
#         R_output = {'R_atm':R_atm,
#                     'R_dir':R_dir,
#                     '_R_dir_coxmunk':R_dir_coxmunk,
#                     '_R_dir_whitecap':R_dir_whitecap,
#                     '_R_dir_water':R_dir_water,
#                     '_R_dir_land':R_dir_land,
#                     'R_env':R_env,
#                     '_R_env_coxmunk':R_env_coxmunk,
#                     '_R_env_whitecap':R_env_whitecap,
#                     '_R_env_water':R_env_water,
#                     '_R_env_land':R_env_land,
#                     'R_total':R_total}
    
#     else:      
#         R_output = {'R_atm':R_atm,
#                     'R_dir':R_dir,
#                     'R_env':R_env,
#                     'R_total':R_total}
    
    
#     return R_output
    
     











# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# %matplotlib qt


# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)

import tmart
import numpy as np
from Py6S.Params.atmosprofile import AtmosProfile

# Specify wavelength in nm
wl = 400

### DEM and reflectance ###

# Three same-size numpy arrays are needed
image_DEM = np.full((2, 2), 0) # in meters
image_reflectance = np.full((2, 2), 0.1) # unitless     
image_isWater = np.full((2, 2), 0) # 1 is water, 0 is land

# pixel width and length, in meters 
cell_size = 20_000 

# Synthesize a surface object
my_surface = tmart.Surface(DEM = image_DEM,
                           reflectance = image_reflectance,
                           isWater = image_isWater,
                           cell_size = cell_size)  

### Atmosphere ###

# Atmophere profile comes from 6S
atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 

# Synthesize an atmosphere object 
my_atm = tmart.Atmosphere(atm_profile)

### Running T-Mart ###

# Make a T-Mart object 
my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm)

# Specify the sensor's position (x, y, z), viewing direction relative 
# to the sensor (zenith, azimuth), sun's direction relative to the target 
# (zenith, azimuth)
my_tmart.set_geometry(sensor_coords=[51,50,130_000], 
                      target_pt_direction=[180,0],
                      sun_dir=[0,0])

# Run and plot on a separate window
%matplotlib qt
results = my_tmart.run_plot(wl=wl, plot_on=True, plot_range=[0,100_000,0,100_000,0,100_000])

# Calculate reflectances using recorded photon information 
R = tmart.calc_ref(results)
for k, v in R.items():
    print(k, '     ' , v)



# This file is part of TMart.
#
# Copyright 2022 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# %matplotlib qt

import sys
sys.path.append('/Users/yw/Desktop/tmart')

import tmart
import numpy as np
import pandas as pd
from Py6S.Params.atmosprofile import AtmosProfile




def calculate_Ediffuse(albedo, ot_abs, ot_rayleigh, sza, n_photon):
    
    # Specify wavelength in nm, not used as atm is specified
    wl = 400
    
    ### DEM and reflectance ###
    
    # Three same-size numpy arrays are needed
    image_DEM = np.full((2, 2), 0) # in meters
    image_reflectance = np.full((2, 2), albedo) # unitless     
    image_isWater = np.full((2, 2), 0) # 1 is water, 0 is land
    
    # pixel width and length, in meters 
    cell_size = 20_000 
    
    # Synthesize a surface object
    my_surface = tmart.Surface(DEM = image_DEM,
                               reflectance = image_reflectance,
                               isWater = image_isWater,
                               cell_size = cell_size)  
    
    ### Atmosphere ###
    
    # Atmophere profile comes from 6S
    atm_profile = AtmosProfile.PredefinedType(AtmosProfile.MidlatitudeSummer) 
    
    # Synthesize an atmosphere object 
    my_atm = tmart.Atmosphere(atm_profile,n_layers=1,
                              specify_abs=ot_abs,
                              specify_ot_rayleigh=ot_rayleigh)
    
    ### Running T-Mart ###
    
    # Make a T-Mart object 
    my_tmart = tmart.Tmart(Surface = my_surface, Atmosphere= my_atm)
    
    # Specify the sensor's position (x, y, z), viewing direction relative 
    # to the sensor (zenith, azimuth), sun's direction relative to the target 
    # (zenith, azimuth), see Parameters for more geometry options
    my_tmart.set_geometry(sensor_coords=[51,50,0.00001], 
                          target_pt_direction='lambertian_up',
                          sun_dir=[sza,0])
    
    results = my_tmart.run(wl=wl,n_photon=n_photon,njobs=2000)
    
    # Calculate reflectances using recorded photon information 
    R = tmart.calc_ref(results, n_photon=n_photon)
    
    print('R: '+str(R))


    return R
    




albedo = 0.1

ot_abs = 0.0

n_photon = 10_000


ot_rayleigh = 0.36
sza = 30
R = calculate_Ediffuse(albedo, ot_abs, ot_rayleigh, sza, n_photon)




'''
results = []

for sza in range(0,100,20):
    
    for ot_rayleigh in np.arange(0.05, 0.55, 0.05):
        print ('\n------------------------------------ SZA: ' + str(sza))
        print ('------------------------------------ ot_rayleigh: ' + str(ot_rayleigh))
        
        R = calculate_Ediffuse(albedo, ot_abs, ot_rayleigh, sza, n_photon)
        results.append({'sza': sza, 'ot_rayleigh': ot_rayleigh, 'E_diffuse': R})

results = pd.DataFrame(results)



results.to_csv('E_diffuse.csv', index=False)

'''































# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Create a masks at different resolutions 

def compute_masks(metadata, config, mask_type):
    # mask_type: 'cloud' or 'all' 
    # output masks in a dictionary, with resolution as keys. E.g., masks['10m'], masks['20m'], etc.
    
    import rasterio, sys, os
    import numpy as np
    
    sensor = metadata['sensor']
    
    # Input from metadata 
    mask_res = metadata['mask_res']
    highTOA_band_names = metadata['highTOA_band_names']
    
    # Number of columns and rows to pad
    pad_columns = metadata['AEC_width'] - metadata['width'] 
    pad_rows = metadata['AEC_height'] - metadata['height'] 
    
    # Mask function 
    def mask_threshold(band_name, threshold, mask_NAN, reshape = True): 
        # Output either an array or a dictionary, mask['10m'], etc. 
        
        # Read file and array 
        band_file = metadata[band_name]
        band_ds = rasterio.open(band_file)
        band_array = band_ds.read(1)
        
        # Resolution of this particular band 
        res_band = int(abs(band_ds.transform[0]))
        
        # Pad less for lower resolution 
        pad_rows_tmp = int(pad_rows/(res_band/metadata['resolution']))
        pad_columns_tmp = int(pad_columns/(res_band/metadata['resolution']))
        
        # Padding 
        band_array = np.pad(band_array, ((0, pad_rows_tmp), (0, pad_columns_tmp)), mode='constant', constant_values=0)
        
        # Scaling  
        scale_mult = metadata[str(band_name) + '_mult']
        scale_add = metadata[str(band_name) + '_add']
        mask_threshold = (float(config[threshold])  - scale_add ) / scale_mult
        
        # Make mask 
        if mask_NAN: mask = np.logical_or(band_array > mask_threshold, band_array == 0)
        else: mask = band_array > mask_threshold
        
        # Add other resolution 
        if reshape:
            
            # Mask 0 as nan 
            band_array = band_array * 1.0
            is_nan = band_array == 0
            band_array[is_nan] = np.nan
            
            masks = {}
            
            # For loop of target resolution  
            for res in mask_res:

                # Height and width at this resolution 
                height_reshaped = int( metadata['AEC_height'] / (res / metadata['resolution'])) 
                width_reshaped = int( metadata['AEC_width'] / (res / metadata['resolution'])) 
    
                # If this is the resolution 
                if res == res_band:
                    masks[str(res) + 'm'] = mask
                    
                # If target resolution is coarser than band resolution, we take the reshaped means 
                elif res > res_band:
    
                    # Suppress warning of mean of empty slice 
                    import warnings
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=RuntimeWarning)
                        band_array_reshaped = band_array.reshape([height_reshaped, int(res / res_band), 
                                                                  width_reshaped, int(res / res_band)])
                        band_array_reshaped = np.nanmean(band_array_reshaped,3)
                        band_array_reshaped = np.nanmean(band_array_reshaped,1)
            
                    if mask_NAN: mask_reshaped = np.logical_or(band_array_reshaped > mask_threshold, np.isnan(band_array_reshaped))
                    else: mask_reshaped = band_array_reshaped > mask_threshold
                        
                    masks[str(res) + 'm'] = mask_reshaped
                
                # If target resolution is finer than band resolution, we repeat cells 
                elif res < res_band: 
                    
                    # stretch to the needed shape, 10980 by 10980 for S2 
                    mask_temp = np.repeat(np.repeat(mask, int(res_band/res), axis=0), int(res_band/res), axis=1)
                    masks[str(res) + 'm'] = mask_temp
                
            return masks
        
        else:
            return mask
    
    # Cirrus band 
    mask_cirrus = mask_threshold(metadata['cirrus_mask'], threshold = 'mask_cirrus_threshold', mask_NAN = False)
    
    # Cloud 
    if mask_type == 'cloud':
        print('Computing cloud mask...')
        
        # Sentinel-2
        if sensor == 'S2A' or sensor == 'S2B':
            gml_file = "{}/MSK_CLOUDS_B00.gml".format(metadata['QI_DATA'])
            jp2_file = "{}/MSK_CLASSI_B00.jp2".format(metadata['QI_DATA'])
            
            # For imagery before processing baseline 4: Jan 25, 2022
            if os.path.exists(gml_file): 
        
                # Built-in cloud mask 
                import geopandas as gpd
                from rasterio.features import geometry_mask
                
                # Load a raster as the base of the mask 
                image = metadata['B02']
                
                with rasterio.open(image) as src:
                    # Read the raster data and transform
                    raster_data = src.read(1)
                    transform = src.transform
                    crs = src.crs
                
                try: 
                    # Read GML file 
                    gdf = gpd.read_file(gml_file)
            
                    # Create a mask using the GML polygons and the GeoTIFF metadata
                    mask_cloud = geometry_mask(gdf['geometry'], transform=transform, out_shape=raster_data.shape, invert=True)
                
                # Sometimes the GML file contains no information, assume no clouds in such case
                except:
                    mask_cloud = np.zeros_like(raster_data)
                
            # For imagery processing baseline 4
            elif os.path.exists(jp2_file): 
                band_ds = rasterio.open(jp2_file)
                band_array = band_ds.read(1)
                mask_cloud = band_array == 1
                mask_cloud = np.repeat(np.repeat(mask_cloud, 6, axis=0), 6, axis=1)
                
            else:
                sys.exit('Warning: cloud mask missing in {}.'.format(metadata['QI_DATA']))
                
        # Landsat 8
        elif sensor == 'L8':
            image = metadata['cloud_mask']
            band_ds = rasterio.open(image)
            band_array = band_ds.read(1)
            
            # Bit operation, source: https://www.usgs.gov/landsat-missions/landsat-collection-1-level-1-quality-assessment-band
            mask_cloud = np.bitwise_and(band_array,2**3) == 2**3
           
        # Padding 
        mask_cloud = np.pad(mask_cloud, ((0, pad_rows), (0, pad_columns)), mode='constant', constant_values=0)
            
        # Overall mask 
        masks = {}
        
        # Original resolution 
        masks[str(metadata['resolution']) + 'm'] = np.logical_or(mask_cirrus[str(metadata['resolution']) + 'm'],mask_cloud)
        
        # Add other resolution 
        for res in mask_res[1:]:
            
            # height and width at this resolution 
            height_reshaped = int( metadata['AEC_height'] / (res / metadata['resolution'])) 
            width_reshaped = int( metadata['AEC_width'] / (res / metadata['resolution'])) 
        
            # Reshape
            mask_cloud_reshaped = mask_cloud.reshape([height_reshaped, int(res / metadata['resolution']), 
                                                      width_reshaped, int(res / metadata['resolution'])]).mean(3).mean(1) > 0
            mask_reshaped = np.logical_or(mask_cirrus[str(res) + 'm'],mask_cloud_reshaped)
            masks[str(res) + 'm'] = mask_reshaped
                
        return masks
    
    # All non-water 
    elif mask_type == 'all':
    
        # SWIR band 
        print('Computing SWIR mask, reading band {}...'.format(config['L8_SWIR_band']))
        mask_SWIR = mask_threshold(metadata['SWIR_mask'],'mask_SWIR_threshold', mask_NAN = True)
        
        ### High TOA bands 
        masks_highTOA = {}
        print('Computing non-water mask, reading bands...')
        for highTOA_band_name in highTOA_band_names:
            print(highTOA_band_name)
            mask_highTOA = mask_threshold(highTOA_band_name,'mask_highTOA_threshold', mask_NAN = True)
            masks_highTOA[highTOA_band_name] = mask_highTOA
            
        # Overall mask 
        masks = {}
        
        # For each resolution 
        for res in mask_res:
            
            # Height and width at this resolution 
            height_reshaped = int( metadata['AEC_height'] / (res / metadata['resolution'])) 
            width_reshaped = int( metadata['AEC_width'] / (res / metadata['resolution'])) 
            
            # Make an empty array for each resolution 
            mask_highTOA_res = np.full((height_reshaped, width_reshaped), False)
            
            # For each highTOA band 
            for highTOA_band_name in highTOA_band_names: 
                mask_highTOA_res = np.logical_or(mask_highTOA_res, masks_highTOA[highTOA_band_name][str(res) + 'm'])
                
            # Overall mask: cirrus, SWIR, and highTOA
            mask_reshaped = np.logical_or(np.logical_or(mask_cirrus[str(res) + 'm'], mask_SWIR[str(res) + 'm']), mask_highTOA_res)
            masks[str(res) + 'm'] = mask_reshaped
                
        return masks   

    else:
        sys.exit('Warning: unrecognized mask was requested')





# This file is part of TMart.
#
# Copyright 2023 Yulun Wu.
#
# TMart is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.


# Read S2 metadata in XML

def read_xml_S2_scene(file):
    
    from xml.dom import minidom
    import sys
    
    # Parse an xml file by name
    xml = minidom.parse(file)
    bands = ['B01','B02','B03','B04','B05','B06','B07','B08','B8A','B09','B10','B11','B12']
    metadata = {}
    tdom = xml.getElementsByTagName('RADIO_ADD_OFFSET')
    
    # Older imagery 
    if len(tdom) == 0:
        for band_name in bands:
            metadata[str(band_name) + '_mult'] = 1/10_000
            metadata[str(band_name) + '_add'] = 0
    
    else: 
        for i in range(len(tdom)):
            sub_tdom = tdom[i]
            band_id = int(sub_tdom.getAttribute('band_id'))
            if band_id == i:
                band_name = bands[i]
                metadata[str(band_name) + '_mult'] = 1/10_000
                metadata[str(band_name) + '_add'] = float(sub_tdom.firstChild.nodeValue) / 10_000
            else: 
                sys.exit('Warning: failed to read {}'.format(file))
    return metadata   
            